﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;

using App4.Dao;
using System.Reflection;
using NHibernate.Criterion;

namespace App4.Utils
{
    public class DataSource
    {
        public string ID { get; set; }
        public string dataFormat { get; set; }
        public ICollection<Dictionary<string, object>> fields { get; set; }
        public ICollection<Dictionary<string, object>> operationBindings { get; set; }

        /// <summary>
        /// Get a field's data by it's name
        /// </summary>
        /// <param name="name">The name of the field</param>
        /// <returns>A Dictionary with the attributes of the field, or null if field was not found</returns>
        public Dictionary<string, object> getField(string name)
        {
            foreach (Dictionary<string, object> field in fields)
            {
                if (field.ContainsKey("name") && field["name"].Equals(name))
                {
                    return field;
                }
            }

            return null;
        }

        virtual public DSResponse execute<T>(DSRequest<T> request)
        {
            if (request == null)
            {
                return null;
            }

            if ("add".Equals(request.operationType))
            {
                return executeAdd<T>(request);
            }

            if ("fetch".Equals(request.operationType))
            {
                return executeFetch<T>(request);
            }

            if ("update".Equals(request.operationType))
            {
                return executeUpdate<T>(request);
            }

            if ("remove".Equals(request.operationType))
            {
                return executeRemove<T>(request);
            }

            return null;
        }

        virtual public DSResponse executeAdd<T>(DSRequest<T> request)
        {
            DSRequest<supplyItem> req = request.RpcManager.convertRequest<supplyItem>(request as DSRequest<Dictionary<string, object>>);

            DataSource ds = request.RpcManager.getDataSource(req.dataSource);

            request.RpcManager.Session.Save(req.data);
            request.RpcManager.Session.Flush();

            DSResponse dsresponse = new DSResponse();

            dsresponse.data = req.data;
            dsresponse.status = 0;

            return dsresponse;
        }

        virtual public DSResponse executeRemove<T>(DSRequest<T> request)
        {
            DSRequest<supplyItem> req = request.RpcManager.convertRequest<supplyItem>(request as DSRequest<Dictionary<string, object>>);

            supplyItem itm = request.RpcManager.Session.Load<supplyItem>(req.data.itemID);

            request.RpcManager.Session.Delete(itm);
            request.RpcManager.Session.Flush();

            DSResponse dsresponse = new DSResponse();

            dsresponse.data = req.data;
            dsresponse.status = 0;

            return dsresponse;
        }

        virtual public DSResponse executeUpdate<T>(DSRequest<T> request)
        {
            DSRequest<Dictionary<string, object>> reqmap = request as DSRequest<Dictionary<string, object>>;
            DSRequest<supplyItem> req = request.RpcManager.convertRequest<supplyItem>(reqmap);

            DataSource ds = request.RpcManager.getDataSource(req.dataSource);

            supplyItem itm = request.RpcManager.Session.Load<supplyItem>(req.data.itemID);

            // update all fields which have changed. they are defined in the data property
            // build the criteria
            if (reqmap.data.Keys.Count != 0)
            {
                foreach (string key in reqmap.data.Keys)
                {
                    PropertyInfo pi = req.data.GetType().GetProperty(key);

                    // get the property value
                    object val = pi.GetValue(req.data, null);

                    // set the property value
                    pi.SetValue(itm, val, null);
                }
            }

            request.RpcManager.Session.Update(itm);
            request.RpcManager.Session.Flush();

            // create a result object to be returned
            // and copy all properties of the updated object into this one
            supplyItem data = new supplyItem();
            if (reqmap.oldValues.Keys.Count != 0)
            {
                foreach (string key in reqmap.oldValues.Keys)
                {
                    PropertyInfo pi = req.data.GetType().GetProperty(key);

                    // get the property value of the updated entity
                    object val = pi.GetValue(itm, null);

                    // store it into the new object
                    pi.SetValue(data, val, null);
                }
            }

            // create the DSResponse object
            DSResponse dsresponse = new DSResponse();
            dsresponse.data = data;
            dsresponse.status = 0;

            return dsresponse;
        }

        virtual public DSResponse executeFetch<T>(DSRequest<T> request)
        {
            DSRequest<Dictionary<string, object>> req = request as DSRequest<Dictionary<string, object>>;
            DSRequest<supplyItem> itmreq = request.RpcManager.convertRequest<supplyItem>(req);

            DataSource ds = request.RpcManager.getDataSource(req.dataSource);

            var query = request.RpcManager.Session.CreateCriteria<supplyItem>();

            // build the criteria
            if (req.data.Keys.Count != 0)
            {
                foreach (string key in req.data.Keys)
                {
                    Dictionary<string, object> field = ds.getField(key);

                    // make sure the field is in the DataSource
                    if (field != null)
                    {
                        // get the property value
                        PropertyInfo pi = itmreq.data.GetType().GetProperty(key);
                        object val = pi.GetValue(itmreq.data, null);

                        string type = field["type"] as string;

                        if (type.Equals("text") || type.Equals("link") || type.Equals("enum") ||
                            type.Equals("image") || type.Equals("ntext"))
                        {
                            query.Add(Restrictions.Like(key, "%" + val + "%"));
                            break;
                        }
                        else
                        {
                            query.Add(Restrictions.Eq(key, val));
                            break;
                        }
                    }
                }
            }

            // add sorting
            if (req.sortBy != null)
            {
                // add the sorting
                foreach (string column in req.sortBy)
                {
                    // if column name is with -, then ordering is descending, otherwise ascending 
                    if (column.StartsWith("-"))
                    {
                        // if sort is descending, then we have to remove the '-' from the beginning to get te correct
                        // column name
                        query.AddOrder(Order.Desc(column.Substring(1)));
                    }
                    else
                    {
                        query.AddOrder(Order.Asc(column));
                    }
                }
            }

            // create a response object
            DSResponse dsresponse = new DSResponse();

            // set start row and number of rows on the query itself
            if (req.endRow != 0)
            {
                query.SetMaxResults(req.endRow - req.startRow);
                query.SetFirstResult(req.startRow);
            }

            // get the requested number of objects
            var products = query.List<supplyItem>();

            // change projection to get the total number of rows
            // we need to clear the result range and the ordering for this to work
            query.SetProjection(Projections.RowCount());
            query.SetFirstResult(0);
            query.SetMaxResults(int.MaxValue);
            query.ClearOrders();

            // set total rows using the projection
            dsresponse.totalRows = (int)query.UniqueResult<int>();

            // set the response data
            dsresponse.data = products;
            dsresponse.startRow = req.startRow;
            dsresponse.endRow = dsresponse.startRow + products.Count();

            // sanity check, if no rows, return 0
            if (dsresponse.endRow < 0)
            {
                dsresponse.endRow = 0;
            }

            dsresponse.status = 0;

            return dsresponse;
        }
    }
}