﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.IO;

using Newtonsoft.Json;
using App4.Dao;
using System.Web.Mvc;
using Newtonsoft.Json.Converters;

using NHibernate;

namespace App4.Utils
{
    public class RPCManager
    {
        HttpRequestBase httpRequest;
        HttpResponseBase httpResponse;
        string requestBody;
        int transactionNum;

        private DSResponse[] responses;
        private string[] requestsAsJSON;
        private DSRequest<Dictionary<string, object>>[] requestsAsMap;

        public ISession Session { get; set; }
        public ITransaction Transaction { get; set; }

        /// <summary>
        /// Create a RPCManager with the given Request and Response objects
        /// </summary>
        /// <param name="request"></param>
        /// <param name="response"></param>
        public RPCManager(HttpRequestBase request, HttpResponseBase response)
        {
            this.httpRequest = request;
            this.httpResponse = response;

            // get request as stream and read the json payload
            var sr = new StreamReader(request.InputStream);
            requestBody = sr.ReadToEnd();
            sr.Close();

            // if is not wrapped in a transaction then we'll wrap it to make unified handling
            // of the request
            if (!hasTransactions())
            {
                requestBody = @"{""transaction"":{""transactionNum"":""-1"",""operations"": [" +
                                       requestBody + @"]}}";
            }
        }

        /// <summary>
        /// Process the request and return back the response
        /// </summary>
        /// <param name="request"></param>
        /// <param name="response"></param>
        /// <returns></returns>
        public static ActionResult processRequest(HttpRequestBase request, HttpResponseBase response)
        {
            RPCManager manager = new RPCManager(request, response);

            return manager.processTransaction();
        }

        /// <summary>
        /// Process the transaction request for which this RPCManager was created for
        /// </summary>
        /// <returns></returns>
        private ActionResult processTransaction()
        {
            // retrieve the requests with data in form of Dictionary<string, object>
            TransactionRequest<Dictionary<string, object>> transactionRequest = parseTransactionRequest<Dictionary<string, object>>();
            requestsAsMap = new DSRequest<Dictionary<string, object>>[transactionRequest.operations.Length];
            transactionRequest.operations.CopyTo(requestsAsMap, 0);
            requestsAsJSON = new string[transactionRequest.operations.Length];
            transactionRequest.strOperations.CopyTo(requestsAsJSON, 0);

            // store transaction num, we'll use it later to see if there was a transaction or not
            transactionNum = transactionRequest.transactionNum;

            // create the array of responses
            responses = new DSResponse[requestsAsMap.Length];

            int idx = 0;
            bool queueFailed = false;

            // create a session and a transaction
            using (ISession session = NHibernateHelper.OpenSession())
            {
                using (var tx = session.BeginTransaction())
                {
                    // store them in RPCManager, so requests can use them
                    Session = session;
                    Transaction = tx;

                    // iterate over the requests
                    foreach (DSRequest<Dictionary<string, object>> req in requestsAsMap)
                    {
                        // set the RPCManager executing this request
                        req.RpcManager = this;

                        //execute the request and get the response
                        DSResponse res = req.execute();

                        // safeguard, if was null, create an empty one with failed status
                        if (res == null)
                        {
                            res = new DSResponse();
                            res.status = -1;
                        }

                        // if request execution failed, mark the flag variable
                        if (res.status == -1)
                        {
                            queueFailed = true;
                        }

                        // store the response for later
                        responses[idx] = res;

                        idx++;
                    }

                    // if there were no errors, commit the transaction
                    if (!queueFailed)
                    {
                        tx.Commit();
                    }
                }
            }

            // if we have only one object, send directly the DSResponse
            if (transactionNum == -1)
            {
                return buildResult(new
                {
                    response = new
                    {
                        status = responses[0].status,
                        startRow = responses[0].startRow,
                        endRow = responses[0].endRow,
                        totalRows = responses[0].totalRows,
                        data = responses[0].data
                    }
                });
            }

            // the list to be sent to front-end
            LinkedList<object> response = new LinkedList<object>();

            // iterate over the responses and create a instance of an anonymous class
            // which mimics the required json
            foreach (DSResponse res in responses)
            {
                response.AddLast(new
                {
                    response = new
                    {
                        status = res.status,
                        startRow = res.startRow,
                        endRow = res.endRow,
                        totalRows = res.totalRows,
                        data = res.data,
                        queueStatus = queueFailed ? -1 : 0
                    }
                });
            }

            return buildResult(response);
        }

        /// <summary>
        /// Returns true if the request has transaction support
        /// </summary>
        /// <returns></returns>
        protected bool hasTransactions()
        {
            if (requestBody.Contains("\"transaction\":") && requestBody.Contains("\"operations\":") && requestBody.Contains("\"transactionNum\":"))
            {
                return true;
            }

            return false;
        }

        /// <summary>
        /// Convert a payload to a TransactionRequest object
        /// </summary>
        /// <typeparam name="T">type of data member of the DSRequest object items</typeparam>
        /// <returns>a TransactionRequest</returns>
        protected TransactionRequest<T> parseTransactionRequest<T>()
        {
            string s = requestBody.Substring(requestBody.IndexOf("\"transaction\":") + "{ transaction:".Length);
            s = s.Substring(0, s.LastIndexOf("}"));

            TransactionRequest<T> req = JsonConvert.DeserializeObject<TransactionRequest<T>>(s);
            req.jsonValue = s;

            s = s.Substring(s.IndexOf("\"operations\": [") + "\"operations\": [".Length);
            s = s.Substring(0, s.LastIndexOf("]}"));
                

            req.strOperations = new LinkedList<string>();

            int pos = 0;
            int block = 0;
            bool instr = false;
            char strsep='\'';
            string request = "";
            while(pos < s.Length)
            {
                if (instr)
                {
                    if (s[pos] == '\\')
                    {
                        request += s[pos++];
                        request += s[pos++];
                    }
                    else if (s[pos] == strsep)
                    {
                        request += s[pos++];

                        instr = false;
                    }
                    else
                    {
                        request += s[pos++];
                    }
                }
                else if (s[pos] == '"')
                {
                    instr = true;

                    strsep = '"';
                    
                    request += s[pos++];
                }
                else if (s[pos] == '\'')
                {
                    instr = true;

                    strsep = '\'';

                    request += s[pos++];
                } 
                else if (s[pos] == '{')
                {
                    request += s[pos++];

                    block++;
                }
                else if (s[pos] == '}')
                {
                    request += s[pos++];

                    block--;

                    if (block == 0)
                    {
                        req.strOperations.Add(request);
                        if (pos < s.Length-1 && s[pos] == ',')
                        {
                            pos++;
                        }

                        request = "";
                    }
                }
                else
                {
                    request += s[pos++];
                }
            }

            return req;
        }

        /// <summary>
        /// Convert one DSRequest to same DSRequest but with internal representation of 
        /// data property as dictionary<string, object>
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        public DSRequest<T> convertRequest<T>(DSRequest<Dictionary<string, object>> request)
        {
            for (int i = 0; i < requestsAsMap.Length; i++)
            {
                if (requestsAsMap[i] == request)
                {
                    DSRequest<T> req = JsonConvert.DeserializeObject<DSRequest<T>>(requestsAsJSON[i]);

                    return req;
                }
            }

            return null;
        }

        /// <summary>
        /// Load a DataSource specified by it's ID
        /// </summary>
        /// <param name="datasourceId">the ID of the DataSource to be loaded</param>
        /// <returns>the DataSource object corresponding to the DataSource with the specified ID or null if not found</returns>
        public DataSource getDataSource(string datasourceId)
        {
            //Load the DataSource. The DataSources are stored in the /ds/ folder in a file
            //with same name as the id of the DataSource
            try
            {
                StreamReader sr = new StreamReader(httpRequest.MapPath("~/ds/" + datasourceId + ".js"));
                var dsStr = sr.ReadToEnd();
                sr.Close();

                dsStr = dsStr.Replace("isc.RestDataSource.create(", "");
                dsStr = dsStr.Replace(");", "");

                DataSource ds = JsonConvert.DeserializeObject<DataSource>(dsStr);

                return ds;
            }
            catch (IOException ex)
            {
                // if not found, return null
                return null;
            }
        }

        /// <summary>
        /// Transforms a object object into a JsonNetResult. Will setup the serializer with the 
        /// appropriate converters, attributes,etc.
        /// </summary>
        /// <param name="dsresponse">the object object to be transformed to JsonNetResult</param>
        /// <returns>the created JsonNetResult object</returns>
        protected JsonNetResult buildResult(object dsresponse)
        {
            // convert it to JSON
            JsonNetResult jsonNetResult = new JsonNetResult();
            jsonNetResult.Formatting = Formatting.Indented;
            jsonNetResult.SerializerSettings.Converters.Add(new IsoDateTimeConverter());
            jsonNetResult.SerializerSettings.NullValueHandling = NullValueHandling.Ignore;

            jsonNetResult.Data = dsresponse;

            return jsonNetResult;
        }

        /// <summary>
        /// Inner class to help deserializing a transaction requet
        /// </summary>
        /// <typeparam name="T"></typeparam>
        protected class TransactionRequest<T>
        {
            public int transactionNum { get; set; }
            public DSRequest<T>[] operations { get; set; }
            public string jsonValue { get; set; }
            public ICollection<string> strOperations { get; set; }
        }
    }
}